/**
 * Copyright (c) 2025 Enzien Audio, Ltd.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions, and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the phrase "powered by heavy",
 *    the heavy logo, and a hyperlink to https://enzienaudio.com, all in a visible
 *    form.
 * 
 *   2.1 If the Application is distributed in a store system (for example,
 *       the Apple "App Store" or "Google Play"), the phrase "powered by heavy"
 *       shall be included in the app description or the copyright text as well as
 *       the in the app itself. The heavy logo will shall be visible in the app
 *       itself as well.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

#include "Heavy_vSwellST.h"
#include "HeavyDPF_vSwellST.hpp"
#include <set>


#define HV_LV2_NUM_PARAMETERS 14

#define HV_HASH_NOTEIN          0x67E37CA3
#define HV_HASH_CTLIN           0x41BE0f9C
#define HV_HASH_POLYTOUCHIN     0xBC530F59
#define HV_HASH_PGMIN           0x2E1EA03D
#define HV_HASH_TOUCHIN         0x553925BD
#define HV_HASH_BENDIN          0x3083F0F7
#define HV_HASH_MIDIIN          0x149631bE
#define HV_HASH_MIDIREALTIMEIN  0x6FFF0BCF

#define HV_HASH_NOTEOUT         0xD1D4AC2
#define HV_HASH_CTLOUT          0xE5e2A040
#define HV_HASH_POLYTOUCHOUT    0xD5ACA9D1
#define HV_HASH_PGMOUT          0x8753E39E
#define HV_HASH_TOUCHOUT        0x476D4387
#define HV_HASH_BENDOUT         0xE8458013
#define HV_HASH_MIDIOUT         0x6511DE55
#define HV_HASH_MIDIOUTPORT     0x165707E4

#define MIDI_RT_CLOCK           0xF8
#define MIDI_RT_START           0xFA
#define MIDI_RT_CONTINUE        0xFB
#define MIDI_RT_STOP            0xFC
#define MIDI_RT_ACTIVESENSE     0xFE
#define MIDI_RT_RESET           0xFF

#define HV_HASH_DPF_BPM         0xDF8C2721

// midi realtime messages
std::set<int> mrtSet {
  MIDI_RT_CLOCK,
  MIDI_RT_START,
  MIDI_RT_CONTINUE,
  MIDI_RT_STOP,
  MIDI_RT_RESET
};


START_NAMESPACE_DISTRHO


// -------------------------------------------------------------------
// Heavy Send and Print hooks

static void hvSendHookFunc(HeavyContextInterface *c, const char *sendName, uint32_t sendHash, const HvMessage *m)
{
  HeavyDPF_vSwellST* plugin = (HeavyDPF_vSwellST*)c->getUserData();
  if (plugin != nullptr)
  {
  }
}

static void hvPrintHookFunc(HeavyContextInterface *c, const char *printLabel, const char *msgString, const HvMessage *m)
{
  char buf[64];
  char* dst = buf;
  int len = strnlen(printLabel, 48);
  dst = strncpy(dst, printLabel, len);
  dst = strcpy(dst, " ");
  dst = strncpy(dst, msgString, 63-len);
  printf("> %s \n", buf);
}

// -------------------------------------------------------------------
// Main DPF plugin class

HeavyDPF_vSwellST::HeavyDPF_vSwellST()
 : Plugin(HV_LV2_NUM_PARAMETERS, 0, 0)
{
  _parameters[0] = 0.6f;
  _parameters[1] = 400.0f;
  _parameters[2] = 0.65f;
  _parameters[3] = 1000.0f;
  _parameters[4] = 2200.0f;
  _parameters[5] = 1.5f;
  _parameters[6] = 1e-07f;
  _parameters[7] = 52.0f;
  _parameters[8] = 1.5f;
  _parameters[9] = 0.0f;
  _parameters[10] = 2.0f;
  _parameters[11] = 0.0f;
  _parameters[12] = 0.0f;
  _parameters[13] = 1.0f;
  

  _context = hv_vSwellST_new_with_options(getSampleRate(), 10, 14, 2);
  _context->setUserData(this);
  _context->setSendHook(&hvSendHookFunc);
  _context->setPrintHook(&hvPrintHookFunc);

  
  // ensure that the new context has the current parameters
  for (int i = 0; i < HV_LV2_NUM_PARAMETERS; ++i) {
    setParameterValue(i, _parameters[i]);
  }
}

HeavyDPF_vSwellST::~HeavyDPF_vSwellST() {
  hv_vSwellST_free(_context);
}

void HeavyDPF_vSwellST::initParameter(uint32_t index, Parameter& parameter)
{// initialise parameters with defaults
  switch (index)
  {
    case paramAttack_Level:
        parameter.name = "Attack Level";
        parameter.symbol = "attack_level";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 3.0f;
        parameter.ranges.def = 0.6f;
        break;
    case paramAttack_ms:
        parameter.name = "Attack ms";
        parameter.symbol = "attack_ms";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 10.0f;
        parameter.ranges.max = 3000.0f;
        parameter.ranges.def = 400.0f;
        break;
    case paramMid_Level:
        parameter.name = "Mid Level";
        parameter.symbol = "mid_level";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 3.0f;
        parameter.ranges.def = 0.65f;
        break;
    case paramMid_ms:
        parameter.name = "Mid ms";
        parameter.symbol = "mid_ms";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 10.0f;
        parameter.ranges.max = 4000.0f;
        parameter.ranges.def = 1000.0f;
        break;
    case paramRelease_ms:
        parameter.name = "Release ms";
        parameter.symbol = "release_ms";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 10.0f;
        parameter.ranges.max = 4000.0f;
        parameter.ranges.def = 2200.0f;
        break;
    case paramShape:
        parameter.name = "Shape";
        parameter.symbol = "shape";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 0.1f;
        parameter.ranges.max = 3.0f;
        parameter.ranges.def = 1.5f;
        break;
    case paramSubEnv_Level:
        parameter.name = "SubEnv Level";
        parameter.symbol = "subenv_level";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 1e-07f;
        parameter.ranges.max = 0.7f;
        parameter.ranges.def = 1e-07f;
        break;
    case paramThreshold_High:
        parameter.name = "Threshold High";
        parameter.symbol = "threshold_high";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 20.0f;
        parameter.ranges.max = 90.0f;
        parameter.ranges.def = 52.0f;
        break;
    case paramThreshold_Low:
        parameter.name = "Threshold Low";
        parameter.symbol = "threshold_low";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 0.3f;
        parameter.ranges.max = 8.0f;
        parameter.ranges.def = 1.5f;
        break;
    case paramThreshold_Strict:
        parameter.name = "Threshold Strict";
        parameter.symbol = "threshold_strict";
        parameter.hints = kParameterIsAutomatable
        | kParameterIsBoolean;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 1.0f;
        parameter.ranges.def = 0.0f;
        break;
    case paramTrigger_Delay:
        parameter.name = "Trigger Delay";
        parameter.symbol = "trigger_delay";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 50.0f;
        parameter.ranges.def = 2.0f;
        break;
    case paramTrigger_EXT:
        parameter.name = "Trigger EXT";
        parameter.symbol = "trigger_ext";
        parameter.hints = kParameterIsAutomatable| kParameterIsTrigger;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 1.0f;
        parameter.ranges.def = 0.0f;
        break;
    case paramTrigger_Left:
        parameter.name = "Trigger Left";
        parameter.symbol = "trigger_left";
        parameter.hints = kParameterIsAutomatable
        | kParameterIsBoolean;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 1.0f;
        parameter.ranges.def = 0.0f;
        break;
    case paramTrigger_Right:
        parameter.name = "Trigger Right";
        parameter.symbol = "trigger_right";
        parameter.hints = kParameterIsAutomatable
        | kParameterIsBoolean;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 1.0f;
        parameter.ranges.def = 1.0f;
        break;
    
  }
  
}

// -------------------------------------------------------------------
// Internal data

float HeavyDPF_vSwellST::getParameterValue(uint32_t index) const
{
  return _parameters[index];
  
}

void HeavyDPF_vSwellST::setParameterValue(uint32_t index, float value)
{
  switch (index) {
    case 0: {
      _context->sendFloatToReceiver(
          Heavy_vSwellST::Parameter::In::ATTACK_LEVEL,
          value);
      break;
    }
    case 1: {
      _context->sendFloatToReceiver(
          Heavy_vSwellST::Parameter::In::ATTACK_MS,
          value);
      break;
    }
    case 2: {
      _context->sendFloatToReceiver(
          Heavy_vSwellST::Parameter::In::MID_LEVEL,
          value);
      break;
    }
    case 3: {
      _context->sendFloatToReceiver(
          Heavy_vSwellST::Parameter::In::MID_MS,
          value);
      break;
    }
    case 4: {
      _context->sendFloatToReceiver(
          Heavy_vSwellST::Parameter::In::RELEASE_MS,
          value);
      break;
    }
    case 5: {
      _context->sendFloatToReceiver(
          Heavy_vSwellST::Parameter::In::SHAPE,
          value);
      break;
    }
    case 6: {
      _context->sendFloatToReceiver(
          Heavy_vSwellST::Parameter::In::SUBENV_LEVEL,
          value);
      break;
    }
    case 7: {
      _context->sendFloatToReceiver(
          Heavy_vSwellST::Parameter::In::THRESHOLD_HIGH,
          value);
      break;
    }
    case 8: {
      _context->sendFloatToReceiver(
          Heavy_vSwellST::Parameter::In::THRESHOLD_LOW,
          value);
      break;
    }
    case 9: {
      _context->sendFloatToReceiver(
          Heavy_vSwellST::Parameter::In::THRESHOLD_STRICT,
          value);
      break;
    }
    case 10: {
      _context->sendFloatToReceiver(
          Heavy_vSwellST::Parameter::In::TRIGGER_DELAY,
          value);
      break;
    }
    case 11: {
      _context->sendFloatToReceiver(
          Heavy_vSwellST::Parameter::In::TRIGGER_EXT,
          value);
      break;
    }
    case 12: {
      _context->sendFloatToReceiver(
          Heavy_vSwellST::Parameter::In::TRIGGER_LEFT,
          value);
      break;
    }
    case 13: {
      _context->sendFloatToReceiver(
          Heavy_vSwellST::Parameter::In::TRIGGER_RIGHT,
          value);
      break;
    }
    
    default: return;
  }
  _parameters[index] = value;
  
}


// -------------------------------------------------------------------
// Process

// void HeavyDPF_vSwellST::activate()
// {

// }

// void HeavyDPF_vSwellST::deactivate()
// {

// }
#if DISTRHO_PLUGIN_WANT_MIDI_INPUT
// -------------------------------------------------------------------
// Midi Input handler

void HeavyDPF_vSwellST::handleMidiInput(uint32_t frames, const MidiEvent* midiEvents, uint32_t midiEventCount)
{
  // Realtime events
  const TimePosition& timePos(getTimePosition());
  bool reset = false;

  if (timePos.playing)
  {
    if (timePos.frame == 0)
    {
      _context->sendMessageToReceiverV(HV_HASH_MIDIREALTIMEIN, 0,
        "ff", (float) MIDI_RT_RESET);
      reset = true;
    }

    if (! this->wasPlaying)
    {
      if (timePos.frame == 0)
      {
        _context->sendMessageToReceiverV(HV_HASH_MIDIREALTIMEIN, 0,
          "ff", (float) MIDI_RT_START);
      }
      if (! reset)
      {
        _context->sendMessageToReceiverV(HV_HASH_MIDIREALTIMEIN, 0,
          "ff", (float) MIDI_RT_CONTINUE);
      }
    }
  }
  else if (this->wasPlaying)
  {
    _context->sendMessageToReceiverV(HV_HASH_MIDIREALTIMEIN, 0,
      "ff", (float) MIDI_RT_STOP);
  }
  this->wasPlaying = timePos.playing;

  // sending clock ticks
  if (timePos.playing && timePos.bbt.valid)
  {
    float samplesPerBeat = 60 * getSampleRate() / timePos.bbt.beatsPerMinute;
    float samplesPerTick = samplesPerBeat / 24.0;

    /* get state */
    double nextClockTick = this->nextClockTick;
    double sampleAtCycleStart = this->sampleAtCycleStart;
    double sampleAtCycleEnd = sampleAtCycleStart + frames;

    if (nextClockTick >= 0 && sampleAtCycleStart >= 0 && sampleAtCycleEnd > sampleAtCycleStart) {
      while (nextClockTick < sampleAtCycleEnd) {
        double delayMs = 1000*(nextClockTick - sampleAtCycleStart)/getSampleRate();
        if (delayMs >= 0.0) {
          _context->sendMessageToReceiverV(HV_HASH_MIDIREALTIMEIN, delayMs,
            "ff", (float) MIDI_RT_CLOCK);
        }
        nextClockTick += samplesPerTick;
      }
    }

    /* save variables for next cycle */
    this->sampleAtCycleStart = sampleAtCycleEnd;
    this->nextClockTick = nextClockTick;
  }

  // Midi events
  for (uint32_t i=0; i < midiEventCount; ++i)
  {
    int status = midiEvents[i].data[0];
    int command = status & 0xF0;
    int channel = status & 0x0F;
    int data1   = midiEvents[i].data[1];
    int data2   = midiEvents[i].data[2];

    // raw [midiin] messages
    int dataSize = *(&midiEvents[i].data + 1) - midiEvents[i].data;

    for (int i = 0; i < dataSize; ++i) {
      _context->sendMessageToReceiverV(HV_HASH_MIDIIN, 1000.0*midiEvents->frame/getSampleRate(), "ff",
        (float) midiEvents[i].data[i],
        (float) channel);
    }

    if(mrtSet.find(status) != mrtSet.end())
    {
      _context->sendMessageToReceiverV(HV_HASH_MIDIREALTIMEIN, 1000.0*midiEvents->frame/getSampleRate(),
        "ff", (float) status);
    }

    // typical midi messages
    switch (command) {
      case 0x80: {  // note off
        _context->sendMessageToReceiverV(HV_HASH_NOTEIN, 1000.0*midiEvents->frame/getSampleRate(), "fff",
          (float) data1, // pitch
          (float) 0, // velocity
          (float) channel);
        break;
      }
      case 0x90: { // note on
        _context->sendMessageToReceiverV(HV_HASH_NOTEIN, 1000.0*midiEvents->frame/getSampleRate(), "fff",
          (float) data1, // pitch
          (float) data2, // velocity
          (float) channel);
        break;
      }
      case 0xA0: { // polyphonic aftertouch
        _context->sendMessageToReceiverV(HV_HASH_POLYTOUCHIN, 1000.0*midiEvents->frame/getSampleRate(), "fff",
          (float) data2, // pressure
          (float) data1, // note
          (float) channel);
        break;
      }
      case 0xB0: { // control change
        _context->sendMessageToReceiverV(HV_HASH_CTLIN, 1000.0*midiEvents->frame/getSampleRate(), "fff",
          (float) data2, // value
          (float) data1, // cc number
          (float) channel);
        break;
      }
      case 0xC0: { // program change
        _context->sendMessageToReceiverV(HV_HASH_PGMIN, 1000.0*midiEvents->frame/getSampleRate(), "ff",
          (float) data1,
          (float) channel);
        break;
      }
      case 0xD0: { // aftertouch
        _context->sendMessageToReceiverV(HV_HASH_TOUCHIN, 1000.0*midiEvents->frame/getSampleRate(), "ff",
          (float) data1,
          (float) channel);
        break;
      }
      case 0xE0: { // pitch bend
        // combine 7bit lsb and msb into 32bit int
        hv_uint32_t value = (((hv_uint32_t) data2) << 7) | ((hv_uint32_t) data1);
        _context->sendMessageToReceiverV(HV_HASH_BENDIN, 1000.0*midiEvents->frame/getSampleRate(), "ff",
          (float) value,
          (float) channel);
        break;
      }
      default: break;
    }
  }
}
#endif




// -------------------------------------------------------------------
// DPF Plugin run() loop

#if DISTRHO_PLUGIN_WANT_MIDI_INPUT
void HeavyDPF_vSwellST::run(const float** inputs, float** outputs, uint32_t frames, const MidiEvent* midiEvents, uint32_t midiEventCount)
{
  handleMidiInput(frames, midiEvents, midiEventCount);
#else
void HeavyDPF_vSwellST::run(const float** inputs, float** outputs, uint32_t frames)
{
#endif
  const TimePosition& timePos(getTimePosition());
  if (timePos.playing && timePos.bbt.valid)
    _context->sendMessageToReceiverV(HV_HASH_DPF_BPM, 0, "f", timePos.bbt.beatsPerMinute);

  _context->process((float**)inputs, outputs, frames);
}

// -------------------------------------------------------------------
// Callbacks

void HeavyDPF_vSwellST::sampleRateChanged(double newSampleRate)
{
  hv_vSwellST_free(_context);

  _context = hv_vSwellST_new_with_options(getSampleRate(), 10, 14, 2);
  _context->setUserData(this);
  _context->setSendHook(&hvSendHookFunc);
  _context->setPrintHook(&hvPrintHookFunc);

  // ensure that the new context has the current parameters
  for (int i = 0; i < HV_LV2_NUM_PARAMETERS; ++i) {
    setParameterValue(i, _parameters[i]);
  }
}


// -----------------------------------------------------------------------
/* Plugin entry point, called by DPF to create a new plugin instance. */

Plugin* createPlugin()
{
    return new HeavyDPF_vSwellST();
}

// -----------------------------------------------------------------------

END_NAMESPACE_DISTRHO